#include <linux/v4l2-subdev.h>
#include "libmegapixels.h"

int
libmegapixels_convert_to_rgb(uint32_t v4l_pixfmt, uint8_t *in_data, long in_size, uint8_t *out_data, long out_size)
{
	switch (v4l_pixfmt) {
		case V4L2_PIX_FMT_SBGGR8:
		case V4L2_PIX_FMT_SGBRG8:
		case V4L2_PIX_FMT_SGRBG8:
		case V4L2_PIX_FMT_SRGGB8:

			break;
		default:
			return 0;
	}
	return 1;
}